package org.dyncle;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import jess.Jesp;
import jess.JessException;
import jess.Rete;


/**
 * Abstract class for Analyze & Decide type component.
 * @version 2.0: Improve to use JSHOP2
 * @author Hiroyuki NAKAGAWA
 */
public abstract class ADTypeComponent extends ComponentBehaviour {
	protected String ruleFileName; //for Jess
	protected Rete engine;//for Jess
	protected PlanList planList; //for JSHOP

	protected Port<?> observePort;

	/**
	 * lowerComps: list of control loops whose priorities are lower than this component.
	 * How to use
	 *   1. Call passivateLowerCLs() to passivate CLs registered in lowerComps.
	 *    (lowerComps is automated calculated by calling passivateLowerCLs() ) 
	 *   2. After execute its own task, call activateLowerCLs() to activate CLs.
	 *   
	 */
	List<ComponentBehaviour> lowerComps; 


	/**
     * This constructor sets the owner agent for this behavior.
     * @param a The agent this behavior belongs to.
     */
	public ADTypeComponent(SASystem a) {
		super(a);
		compType = ComponentType.ANALYZE_AND_DECIDE;
		//create rule engine
		engine = new Rete(); 		//for Jess
		planList = new PlanList(); //for JSHOP
	}

	public ADTypeComponent(SASystem a, String name, String serviceType){
		super(a, name, serviceType);
		compType = ComponentType.ANALYZE_AND_DECIDE;
		//create rule engine
		engine = new Rete(); 		//for Jess
		planList = new PlanList(); //for JSHOP
	}


	public Rete getEngine() {
		return engine;
	}

//	public <T> void setObservePort(){
//		observePort = new Port<T>("ObserveResult",PortType.REQUIRED);
//	}

	/**
	 * Reads rules and sets them into the Jess planning engine.
	 * @param ruleFile
	 * @throws JessException 
	 */
	public void setRules(String ruleFile) throws JessException {
		this.ruleFileName = ruleFile;
		
		FileReader fr=null;
		try {
			fr = new FileReader(ruleFile);
			//create a parser for the file
			Jesp jesp = new Jesp(fr, engine);
			jesp.parse(false);
		} catch (FileNotFoundException e) {
			System.err.println("Cannot open rule file: "+this.ruleFileName);
			e.printStackTrace();
//		} catch (JessException e) {
//			e.printStackTrace();
		} finally{
			try {
				fr.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}    

	/**
	 * Generates plans.
	 * @param scriptFineName: script file for start up JSHOP
	 */
	public void makePlan(String scriptFile){
		List<String> result = new ArrayList<String>();
		CygwinExecution cygwinExe = new CygwinExecution();	//Cygwin test
		List<String> list = new ArrayList<String>();
		//describe shell script file to be executed
		list.add(scriptFile);
//		list.add("sh/jshop/jshop2cleaning.sh");
////		list.add("/cygdrive/c/nakagawa/eclipse/workspace/TestEnvironment/resources/jshop/jshop2cleaning.sh");
////		list.add("/cygdrive/c/nakagawa/01UEC/home/sh/wget.sh");
		try {
			result = cygwinExe.execute(list);
		} catch (Exception e1) {
			e1.printStackTrace();
		}
		
		//TODO: Introduce Plan list that register Plan class (manages plan name, arg1, arg2).
		//Parse the result (Extract actions that start by "(!" , except "(!!")
		for(int i=0;i<result.size();i++){
			String line = result.get(i);
			System.out.println(line);
			if(line.startsWith("(!") && !line.startsWith("(!!")){
				String[] words = line.split(" ");
//				for(String st: words){
//					System.out.println("\t\t"+st);
//				}
				StringBuffer sb = new StringBuffer();
				int j=2; //for check next to "(!"
				while(j<words[0].length() && words[0].charAt(j)!=')'){
					sb.append(words[0].charAt(j++));
				}
//				System.out.println("\t\t\tAction name: "+sb.toString());
				Plan plan = new Plan(sb.toString());
				
				if(2<=words.length){ //arg1 exists
					j=0;
					sb = new StringBuffer();
					while(j<words[1].length() && words[1].charAt(j)!=')'){
						sb.append(words[1].charAt(j++));
					}
//					System.out.println("\t\t\tArg1: "+sb.toString());
					plan.setArg1(sb.toString());
					
					if(3<=words.length){ //arg2 exists
						j=0;
						sb = new StringBuffer();
						while(j<words[2].length() && words[2].charAt(j)!=')'){
							sb.append(words[2].charAt(j++));
						}
//						System.out.println("\t\t\tArg2: "+sb.toString());
						plan.setArg2(sb.toString());
					}
				}
//				System.out.println("\tThis line contains action.");
				planList.addList(plan);
			}
		}

	}


	
	/**
	 * Passivates control loops whose priorities are lower than this component.
	 * These control loops are registered in "lowerComps", 
	 * but not active control loops are outside the scope.
	 */
	public void passivateLowerCLs(){
		lowerComps = myAgent.getLowerPriorityCLs(this.name);
		for (Iterator<ComponentBehaviour> iterator = lowerComps.iterator(); iterator.hasNext();) {
			ComponentBehaviour cb = iterator.next();

//		for(ComponentBehaviour cb: lowerComps){
			if(cb.mode!=Mode.ACTIVE){//not active -> delete from the list
//				lowerComps.remove(cb);
				iterator.remove();
			}
			else{//Active
				System.out.println("Lower priority component "+cb.getName()+ " will be passivate.");
				cb.passivate();
			}
		}
	}

	/**
	 * Activates CLs that are previously passivated.
	 */
	public void activateLowerCLs(){
		for(ComponentBehaviour cb: lowerComps){
			System.out.println("Lower priority component "+cb.getName()+ " will be acctivate.");
			cb.activate();
		}

	}

}
