package org.dyncle;

import jade.core.behaviours.SimpleBehaviour;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * A behavior corresponding to a component in design phase.
 * @version 1.4: add "priority"
 * @author Hiroyuki NAKAGAWA
 */
public abstract class ComponentBehaviour extends SimpleBehaviour {
	
	public Mode mode;
	public String name;
	public String serviceType;
	public int priority; //strong:0 -  default:4  - weak
	private final int _MEDIUM_ = 4;
	private boolean isTerminated;
	protected ArrayList<Port<?>> ports;
	protected SASystem myAgent;
	protected ComponentType compType;
	protected boolean isExclusive; //Whether the component provides exclusive service

	/**
     * This constructor sets the owner agent for this behavior.
     * @param a The agent this behavior belongs to.
     */
	public ComponentBehaviour(SASystem a) {
		super(a);
		myAgent = a;
		ports = new ArrayList<Port<?>>();
		isTerminated = false;
		priority = _MEDIUM_;
		this.mode = Mode.WAITING;
		isExclusive = true;
	}    

	/**
	 * This constructor sets the component's name and service type.
	 * @param a The agent this behavior belongs to.
	 * @param name component's name.
	 * @param serviceType components's service type.
	 */
	public ComponentBehaviour(SASystem a, String name, String serviceType) {
		this(a);
		this.name = name;
		this.serviceType = serviceType;
	}    


	@Override
	public final void action() {
		if(mode!=Mode.ACTIVE){
			if(activateCondition()) mode = Mode.ACTIVE;
		}
		if(mode==Mode.ACTIVE){
			perform();
		}
	}
	
	public abstract void perform();

	@Override
	public boolean done() {
//		if(mode==Mode.ACHIEVED || mode==Mode.NOT_ACHIEVED) return true;
//		else return isTerminated;
		return isTerminated;
	}

	public void activate(){
		mode = Mode.ACTIVE;
	}
	
	public void passivate(){
		mode = Mode.WAITING;
	}
	
	public void terminate(){
		isTerminated = true;
		for(Port<?> p: ports){
			System.out.println("Disconnect the port: "+ p.getName());
			if(p.getType()==PortType.REQUIRED){
				disconnectPorts(p, p.getPairPort());
			}
			if(p.getType()==PortType.PROVIDED){
				disconnectPorts(p.getPairPort(), p);
			}
		}
		
	}
	
	protected boolean activateCondition(){
		return false;
	}

	
	public String getName() {
		return name;
	}
	public void setName(String name) {
		this.name = name;
	}
	public String getServiceType() {
		return serviceType;
	}
	public void setServiceType(String serviceType) {
		this.serviceType = serviceType;
	}
	
	/**
	 * Return priority of this component.
	 * This priority corresponds to that of the control loop.
	 * @return
	 */
	public int getPriority(){
		//TODO: add code for components except AD type
		CLInfo clInfo = myAgent.clTable.getCLInfo(this.name);
		if(clInfo==null){
			System.err.println("ComponentBehaviour.getPriority: "+this.name +
					" does not exist in control loops");
			return -1;
		}
		else{
			return clInfo.getPriority();
		}
	}

	//////////////////////////////////// Port ////////////////////////////////////
	public void addPort(Port<?> port){
		this.ports.add(port);
	}
		
	/**
	 * Connects two ports.
	 * If one provides exclusive service, it determines 
	 * whether it should connect or not by comparing the priorities.
	 * @param p1 own port (required Port)
	 * @param p2@partner's port (provided Port)
	 */
	public void connectPorts(Port<?> p1, Port<?> p2){
		if(isExclusive){
			Port<?> existingPairPort;
			if(p2.getType()==PortType.PROVIDED && 
					(existingPairPort = p2.getPairPort())!=null){
				//compare its priority with that of current connected component
				if(existingPairPort.getPriority()<=p1.getPriority()){
					//current connected component has higher priority -> Its port can't be connected
					return;
				}
				else{
					//its own port has higher priority -> Disconnect and establish new connection
					disconnectPorts(existingPairPort, p2);
				}
			}
			
		}
//		System.out.println("Connect between "+p1.getComponent().getName()+
//				" and "+p2.getComponent().getName()+".");
		p1.setPairPort(p2);
		p2.setPairPort(p1);
	}

	
	/**
	 * Disconnects ports.
	 * @param p1 own port (required Port)
	 * @param p2@partner's port (provided Port)
	 */
	public void disconnectPorts(Port<?> p1, Port<?> p2){
		if(p1!=null && p2!=null && 
				p1.getPairPort()==p2 && p2.getPairPort()==p1){
//			System.out.println("Disconnect between "+p1.getComponent().getName()+
//					" and "+p2.getComponent().getName()+".");
			if(isExclusive){
				//Passivate the partner (Initialize the service)
				//TODO: This task should be executed by itself.
				p2.getComponent().passivate();
			}
			p1.setPairPort(null);
			p2.setPairPort(null);
		}
		else{
//			System.err.println("ComponentBehaviour.disconnectPorts(): Port mismatch error.");
			String p1name = null;
			String p1CompName = null;
			String p2name = null;
			String p2CompName = null;
			if(p1!=null){
				p1name = p1.getName();
				p1CompName = p1.getComponent().getName();
			}
			if(p2!=null){
				p2name = p2.getName();
				p2CompName = p2.getComponent().getName();
			}
			System.err.println("ComponentBehaviour.disconnectPorts(): Port mismatch error.\n"+
					"\tp1: "+p1CompName+"."+p1name+"\tp2: "+p2CompName+"."+p2name);
		}
	}
	
	public ArrayList<Port<?>> getRequiredPorts() {
		ArrayList<Port<?>> list = new ArrayList<Port<?>>();
		for(Iterator<Port<?>> i = ports.iterator();i.hasNext();){
			Port<?> p= (Port<?>)i.next();
			if(p.getType()==PortType.REQUIRED){
				list.add(p);
			}
		}
		return list;
	}

	public ArrayList<Port<?>> getProvidedPorts() {
		ArrayList<Port<?>> list = new ArrayList<Port<?>>();
		for(Iterator<Port<?>> i = ports.iterator();i.hasNext();){
			Port<?> p= (Port<?>)i.next();
			if(p.getType()==PortType.PROVIDED){
				list.add(p);
			}
		}
		return list;
	}
}
