package org.dyncle;

import jade.content.lang.sl.SLCodec;
import jade.core.Agent;
import jade.core.behaviours.CyclicBehaviour;
import jade.domain.FIPANames;
import jade.lang.acl.ACLMessage;

import java.lang.ref.WeakReference;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.cli.BasicParser;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;


/**
 * An Agent class for implementing self-adaptive systems.
 * @version 3.2
 * @author Hiroyuki NAKAGAWA
 */
public abstract class SASystem extends Agent {
	
   private final static boolean isDebugMode = true;

	protected Map<String, List<ComponentBehaviour>> componentRepository;

	/** Common data space(HashMap) */
	protected Map<String, Object> dataTable;
	
	/** HashMap for storing classloader  */
	public Map<String, CLLoader> classLoaderRepository;

	/** A reference for checking whether the target has been unloaded  */
	public WeakReference<Object> clWR;
	
	/** Control loop management table. <Mode name -> AD component of control loop> */
	protected CLTable clTable;

	/**
     * Default constructor. It does not set the owner agent for this
     * behaviour.
	 */
	public SASystem(){
		super();
		dataTable = new IgnoreCaseHashMap<String, Object>();
		componentRepository = new IgnoreCaseHashMap<String, List<ComponentBehaviour>>();
		clTable = new CLTable();
		classLoaderRepository = new IgnoreCaseHashMap<String, CLLoader>();
	}

	/**
	 * Sets up the agent.
	 */
	protected void setup() {
		try {
			super.setup();

			addBehaviour(new ReceiveCommand(this));
			
	      this.getContentManager().registerLanguage(new SLCodec(), FIPANames.ContentLanguage.FIPA_SL);
//	      this.getContentManager().registerOntology(EvolutionOntology.getInstance(), EvolutionOntology._ONTOLOGY_NAME);

		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Stands by the component. (= addBehaviour() in JADE)
	 * @param comp
	 */
	public void standBy(ComponentBehaviour comp){
		comp.passivate();
		addBehaviour(comp);
	}

	/**
	 * Registers the specified component.
	 * @param serviceType : Service type (Service name)
	 * @param compB : Component to be registered
	 */
	public void registerComponent(String serviceType, ComponentBehaviour compB){
		List<ComponentBehaviour> compList = componentRepository.get(serviceType);
		if(compList==null){
			compList = new ArrayList<ComponentBehaviour>();
			componentRepository.put(serviceType, compList);
		}
		compList.add(compB);
	}
	
	public List<ComponentBehaviour> findService(String serviceType){
		return componentRepository.get(serviceType);
	}

	/**
	 * Returns a component searched by name.
	 * This search ignores case (not case sensitive).
	 * @param targetCompName : Comonent name.
	 * @return A component whose name is targetCompName
	 */
	public ComponentBehaviour findComponent(String targetCompName){
		Set<String> map_key = componentRepository.keySet();
		for(Iterator<String> i= map_key.iterator(); i.hasNext();){
			List<ComponentBehaviour> compList = componentRepository.get((String)i.next());
			for(Iterator<ComponentBehaviour> j= compList.iterator(); j.hasNext();){
				ComponentBehaviour tmpComp= j.next();
				if(tmpComp.name!=null && tmpComp.name.toLowerCase().equals(targetCompName.toLowerCase())){
					return tmpComp;
				}
			}
		}
		return null;
	}
	
	/**
	 * Returns components that provide same service of specified component, except specified component itself.
	 * (Returns components that has same serviceType)
	 * @param compB
	 * @return
	 */
	public List<ComponentBehaviour> findAnotherService(ComponentBehaviour compB){
		String serviceType = compB.serviceType;
		List<ComponentBehaviour> returnCompList = new ArrayList<ComponentBehaviour>();
		List<ComponentBehaviour> allCompList = componentRepository.get(serviceType);
		for(Iterator<ComponentBehaviour> i= allCompList.iterator(); i.hasNext();){
			ComponentBehaviour tmpComp= i.next();
			if(!tmpComp.name.equals(compB.name)){
				returnCompList.add(tmpComp);
			}
		}
		return returnCompList;
	}
	
	
	////////////////////////////////////Dynamic deployment ////////////////////////////////////
	/**
	 * A behaviour class that recieve deployment commands. <br/>
	 * @author Hiroyuki NAKAGAWA
	 */
	class ReceiveCommand extends CyclicBehaviour {

		public ReceiveCommand(SASystem agent) {
			super(agent);
		}

		/**
		 * Execution task.
		 */
		public void action() {
			try {
            
				ACLMessage receive = receive();

				if(receive != null) {
            	System.out.println("Receive request!!!!");
//               printReceivedCommand(receive);
               // Parse commands by using Apache Jakarta Commons CLI

               String inputCommand = receive.getContent();
               System.out.println("Simple Content: "+receive.getContent());
               String[] dividedCommand = inputCommand.split(" ");
//               System.out.println("Divided command");
//               for(String st: dividedCommand){
//               	System.out.println("\t"+st);
//               }

               String mainCommand = dividedCommand[0];
               String[] optionStrings = Arrays.copyOfRange(dividedCommand, 1, dividedCommand.length);
//               System.out.println("Options");
//               for(String st: optionStrings){
//               	System.out.println("\t"+st);
//               }
               
//               int count;
//               boolean isDebug;
               String clName = null;
               String fqn = null;
            	String classpath = null;
            	String[] confs = null;
               int priority = 100; //Tentative large value (priority: low)
            	List<String> confResourcesList = null;


               Options options = new Options();
//            	options.addOption("c", true, "count number");
//            	options.addOption("d", false, "debug option");
            	options.addOption("p", true, "Priority");
            	options.addOption("n", true, "CL name");
            	options.addOption("fqn", true, "FQN (Fully Qualified Name)");
            	options.addOption("path", true, "classpath");
            	options.addOption("conf", true, "conflict resources");

            	CommandLineParser parser = new BasicParser();
            	CommandLine commandLine;
            	try {
            	    commandLine = parser.parse(options, dividedCommand);
            	} catch (ParseException e) {
            	    System.err.println("Argument parser error");
            	    return;
            	}

            	if (commandLine.hasOption("n")) {
            		clName = commandLine.getOptionValue("n");
            		System.out.println("CL name: "+clName);
            	}
            	if (commandLine.hasOption("fqn")) {
            		fqn = commandLine.getOptionValue("fqn");
            		System.out.println("FQN: "+fqn);
            	}
            	if (commandLine.hasOption("path")) {
            		classpath = commandLine.getOptionValue("path");
            		System.out.println("classpath: "+classpath);
            	}
            	if (commandLine.hasOption("p")) {
            		priority = Integer.parseInt(commandLine.getOptionValue("p"));
            		System.out.println("Priority: "+priority);
            	}
            	if (commandLine.hasOption("conf")) {
            		confResourcesList = new ArrayList<String>();
            		System.out.println("Conflict resources: ");
            		confs = commandLine.getOptionValue("conf").split(",");
            		for(String conf: confs){
            			System.out.print(conf+"\t");
            			confResourcesList.add(conf);
            		}
            		System.out.println();
            	}

            	ComponentBehaviour cl=null;

            	///////////////// Tasks for each command /////////////////
            	if(mainCommand.equals("addCL")){
            		System.out.println("++++ ADD CL: "+clName+" ++++");
               	
            		if(classpath!=null){
            			//Launch control loops using their "own" classloaders.
            			CLLoader ucLoader = new CLLoader(classpath);
            			cl = ucLoader.loadCL(fqn, (SASystem)myAgent);
            		}
            		else{
            			//Launch control loops using "System Classloader".
            			// (The control loops can't be unloaded.)
							try {
								Class<?>[] classes ={SASystem.class};
//								Class<?>[] classes ={CleaningRobot.class};
								Constructor<?> constructor = Class.forName(fqn).getConstructor(classes);
								Object[] param = {myAgent};
								cl = (ComponentBehaviour)constructor.newInstance(param);
							} catch (Exception e) {
								e.printStackTrace();
								
								// getCause(): return Throwable object that causes the exception.
				            Throwable throwable = e.getCause();
				            throwable.printStackTrace();
				  
							}
            		}
            		if(cl!=null){
                  	addBehaviour(cl);
//                  	cl.activate();
                  	//register to CLTable
							//Tentatively register CL with the weakest priority.
                  	clTable.addCLInfo(clName, fqn, cl, Integer.MAX_VALUE, confResourcesList);
                  	if(priority!=0){
                  		CLInfo clInfo = clTable.getCLInfo(clName);
                  		clInfo.setPriority(priority);
                  	}

                  	clTable.printAllCLs();
            		}
            		else{//Specified CL can't be loaded
            			System.err.println("Error in addCL: "+clName+" can't be loaded.");
            		}
               	
               	block();
            	}

            	if(mainCommand.equals("rmCL")){
            		System.out.println("++++ REMOVE CL: "+clName+" ++++");
               	
            		//terminate CL
            		cl = getCL(clName);
            		cl.passivate();
            		cl.terminate();
            		removeBehaviour(cl);

               	//delete CL from CLTable
               	clTable.removeCLInfo(clName);
               	clTable.printAllCLs();

               	//Unload CL
                	
//           	    WeakReference<Object> clWR = new WeakReference<Object>(cl);
               	clWR = new WeakReference<Object>(cl);
         	    if(clWR.get()!=null){
//         	   	 System.out.println("cl is still reading...");
         	    }
         	    else{
         	   	 System.out.println("cl has already been unloaded.");
         	    }
         	    
         	    //set classloader for unload
         	    cl = null;
         	    CLLoader loader = classLoaderRepository.get(clName);
         	    classLoaderRepository.remove(clName);
         	    if(loader!=null){
            	    loader.remove();
            	    loader = null;
         	    }
         	    System.gc();

         	    if(clWR.get()!=null){
//         	   	 System.out.println("cl is still reading...");

         	   	 //create a behavior to delete the control loop because it can't be deleted immediately.
//         	   	 RemoveCL removeCL = new RemoveCL(clWR);
         	   	 RemoveCL removeCL = new RemoveCL((SASystem)myAgent);
         	   	 addBehaviour(removeCL);
         	    }
         	    else{
         	   	 System.out.println("cl has already been unloaded.");
         	    }

               	
               	block();
            	}
            	
            	
            	if(mainCommand.equals("setPri")){	//Change priority
               	System.out.println("++++ SET PRIORITY: "+clName+" ++++");

               	//Register to CLTable
               	CLInfo clInfo = clTable.getCLInfo(clName);
               	if(clInfo==null){
               		System.out.println("Evolution ERROR: Can't find the control loop "+clName);
               	}
               	else{
                  	clInfo.setPriority(priority);
               	}

               	clTable.printAllCLs();
               	
               	block();
            	}
            	
            	if(mainCommand.equals("activateCL")){	//Activate control loop
               	System.out.println("++++ ACTIVATE CL: "+clName+" ++++");

               	clTable.getCLInfo(clName).getRootComp().activate();
               	
               	block();
            	}
            	
            	if(mainCommand.equals("passivateCL")){	//Passivate control loop
               	System.out.println("++++ PASSIVATE CL: "+clName+" ++++");

               	clTable.getCLInfo(clName).getRootComp().passivate();
               	
               	block();
            	}
            	
            	if(mainCommand.equals("addConf")){ //Add conflict target
            		System.out.println("++++ ADD Conflict objects: "+clName+" ++++");
               	
            		List<String> confList = clTable.getCLInfo(clName).getConfResources();
            		for(int i=0; i<confs.length; i++){
            			boolean isResigtered = false; //Check whether the target has already been registered
            			String conf = confs[i];
            			for(int j=0; j<confList.size();j++){
            				if(confList.get(j).equals(conf)){
            					isResigtered = true;
            				}
            			}
            			if(isResigtered){
            				System.out.println("\tConflict object "+conf+" has been already resigtered.");
            			}
            			else{
                  		confList.add(confs[i]);
            			}
            		}
               	clTable.printAllCLs();
               	
               	block();
            	}

            	if(mainCommand.equals("rmConf")){ //Remove conflict target
            		System.out.println("++++ REMOVE Conflict objects: "+clName+" ++++");
               	
            		List<String> confList = clTable.getCLInfo(clName).getConfResources();
            		for(int i=0; i<confs.length; i++){
            			String conf = confs[i];
            			for(int j=0; j<confList.size();j++){
            				if(confList.get(j).equals(conf)){
            					System.out.println("\tRemove conflict object "+conf);
            					confList.remove(j);
            				}
            			}
            		}
               	clTable.printAllCLs();
               	
               	block();
            	}

            } else {
               block();
            }
         } catch (Exception e) {
            e.printStackTrace();
         }
      }
   }
	


	/**
    * Prints received command.
    * @param receive
    */
   public void printReceivedCommand(ACLMessage receive) {
      if(isDebugMode){
         System.out.println(getLocalName() + " got a command from " + receive.getSender().getLocalName() + ">>" + receive);
      }
   }
   
//////////////////////////////////// Methods for CL management ////////////////////////////////////

   public CLTable getCLTable(){
   	return clTable;
   }
   
   /**
    * Adds the data of specified CL into the CLTable, which is a CL management table.
    * @param name
    * @param fqn
    * @param cb
    * @param pri
    * @param confList
    */
	public void addCLInfo(String name, String fqn, ComponentBehaviour cb, 
			int pri, List<String> confList){
		clTable.addCLInfo(name, fqn, cb, pri, confList);
	}

   /**
    * Returns the AD type component of specified CL.
    * @param clName : name of the specified Control loop.
    * @return AD type component of the control loop.
    */
   public ComponentBehaviour getCL(String clName){
   	if(clTable.getCLInfo(clName)==null){
   		return null;
   	}
   	else{
      	return clTable.getCLInfo(clName).getRootComp();
   	}
   }
   
   /**
    * Checks whether the specified CL can transit to the activate state or not.
    * If the other control loop with higher priority is running, 
    * the control loop can't transit to the activate state.
	 * @param name
	 * @return@true if the specified CL can transit to the activate state.
    */
   public boolean canActivate(String name){
   	return clTable.canActivate(name);
   }
   
	/**
	 * Finds the control loops whose priority is lower than that of specified control loop.
	 * @param name : control loop name. 
	 * @return@AD type components whose control loops have lower priority than the specified control loops.
	 */
	public List<ComponentBehaviour> getLowerPriorityCLs(String name){
		return clTable.getLowerPriorityCLs(name);
	}
}
