package robot.de.comp;

import java.awt.Point;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Set;

import org.dyncle.ComponentBehaviour;
import org.dyncle.Mode;
import org.dyncle.Port;
import org.dyncle.PortException;
import org.dyncle.PortType;
import org.nak.cleaning.Direction;
import org.nak.cleaning.Simulator;
import org.nak.cleaning.VirtualRobot.BatteryEmptyException;

import robot.de.CleaningRobot;

/**
 * Component for finding target by random walk <br/>
 * @author Hiroyuki NAKAGAWA
 * @since 2010.5.31
 */
public class DiscoverObjectByRandomWalk extends ComponentBehaviour {
	String targetType; //type of target object (e.g., "Dust", "ChargeStation")
	Point destination;

	// temporary map that is used when the camera module is broken
	// (Coordinates of the map are difference from those of the actual map)
	HashMap<Point,IdentifiedObject> memorizedMap;

	enum IdentifiedObject{NONE, DUST, STATION, OUT_OF_BOUND};

	DetermineDestination determineDestination = null;
	MoveWithMemorization moveWithMemorization = null;
	Port<Point> determineDestinationPort;
	Port<Point> moveWithMemorizationPort;
	Port<Point> moveToPort;

	Simulator simulator;

	// temporary variables for storing current estimated point of the robot
	// It is used when the camera module is broken.
	// (values are difference from the actual location)
	Point currentEstimatedPoint;
	
	// temporary variables for storing max and min coordinates of the map
	// It is used when the camera module is broken.
	// (values are difference from the actual map data)
	int fixed_x_min;
	int fixed_x_max;
	int fixed_y_min;
	int fixed_y_max;

	
	// when this component is activate,  it prints out the component name just once
	boolean forPrint=false;

	public boolean done(){
		return false;
	}

	/**
	 * Constructor
	 * @param eType specified object type
	 */
	public DiscoverObjectByRandomWalk(CleaningRobot agent, String eType) {
		super(agent);
		mode = Mode.ACTIVE;
		determineDestinationPort = new Port<Point>("TargetPoint",PortType.REQUIRED,this);
		moveWithMemorizationPort = new Port<Point>("DestinationPoint",PortType.REQUIRED,this);
		
		moveWithMemorization = new MoveWithMemorization((CleaningRobot)myAgent);
		connectPorts(moveWithMemorizationPort, moveWithMemorization.getProvidedPorts().get(0));
		myAgent.standBy(moveWithMemorization);
		moveWithMemorization.passivate();
		simulator =((CleaningRobot)myAgent).simulator;

		forPrint = true;
		destination = null;
		targetType = eType;
	}

	/**
	 * Execution task
	 */
	public void perform() {
		if(forPrint){
			System.out.println("Component: RondomLyApproach");
			forPrint = false;
		}
		if(memorizedMap==null){//clear memorized map at the first time
			memorizedMap = new HashMap<Point, IdentifiedObject>();
			if(!simulator.existsDust())	memorizedMap.put(new Point(0,0), IdentifiedObject.NONE);
			currentEstimatedPoint = new Point(0,0);
			fixed_x_min = Integer.MIN_VALUE;
			fixed_x_max = Integer.MAX_VALUE;
			fixed_y_min = Integer.MIN_VALUE;
			fixed_y_max = Integer.MAX_VALUE;

		}
		boolean isMoveRequired=true;
		if(targetType.equals("ChargeStation")){
			if(simulator.isLocatedOnStation()){
				isMoveRequired = false;
				destination = currentEstimatedPoint; //no need to move
			}
		}
		if(targetType.equals("Dust")){
			if(simulator.existsDust()){
				isMoveRequired = false;
				destination = currentEstimatedPoint; //no need to move
			}
		}

		if(!isMoveRequired){ //already arrived
			mode = Mode.ACHIEVED;
		}
		else{
			if(determineDestination == null){
				determineDestination = new DetermineDestination((CleaningRobot)myAgent, targetType);
				connectPorts(determineDestinationPort, determineDestination.getProvidedPorts().get(0));
				myAgent.standBy(determineDestination);
			}
			if(determineDestination.mode==Mode.ACHIEVED){
//				System.out.println("cccc");
				if(moveWithMemorization.mode==Mode.WAITING){
					try {
						destination = determineDestinationPort.get();
						if(destination!=null){
							System.out.println("next target point: "+destination);
							moveWithMemorizationPort.put(destination);
							moveWithMemorization.activate();
						}
					} catch (PortException e) {
						e.printStackTrace();
					}
				}
				else{
					if(moveWithMemorization.mode == Mode.ACHIEVED){
						System.out.println("reached at "+destination);
						if((targetType.equals("Dust") && simulator.existsDust()) ||
								(targetType.equals("ChargeStation") && simulator.isLocatedOnStation())){
							mode = Mode.ACHIEVED;
						}
						determineDestination = null;
						moveWithMemorization.passivate();
					}
					else{
						if(moveWithMemorization.mode == Mode.NOT_ACHIEVED){
							// select destination again
							determineDestination = null;
							moveWithMemorization.passivate();
						}
					}
				}
			}
		}
	}

	/** Activates the component's service */
	public void activate(){
		super.activate();
		forPrint = true;
	}
	
	/**
	 * Finds nearest object whose type is specified. <br/>
	 * outputF the nearest point that the specified object exists
	 * @author Hiroyuki NAKAGAWA
	 */
	class DetermineDestination extends ComponentBehaviour {
		String targetType;
		boolean isFinished;
		Port<Point> outputPort;

		public boolean done(){
			return isFinished;
		}

		/**
		 * Constructor
		 * @param eType specified object type
		 */
		 public DetermineDestination(CleaningRobot agent, String eType) {
			super(agent);
			outputPort = new Port<Point>("TargetPoint", PortType.PROVIDED,this);
			this.addPort(outputPort);
			mode = Mode.ACTIVE;
			this.targetType = eType;
		}

			/**
			 * Execution task
			 */
			public void perform() {
				Point targetPoint=null;
				System.out.println("Component: DetermineDestination");
				
				if(targetType.equals("ChargeStation")){
					//look for the station place from the random walk result until now
					targetPoint=getNearestPoint(currentEstimatedPoint, IdentifiedObject.STATION);
				}

				if(targetType.equals("Dust")){
					//look for the dust item place from the random walk result until now
					if(simulator.watchObject()==null){//nothing
						memorizedMap.put(new Point(currentEstimatedPoint.x, currentEstimatedPoint.y), 
								IdentifiedObject.NONE);
					}
					targetPoint=getNearestPoint(currentEstimatedPoint,IdentifiedObject.DUST);
				}

				if(targetPoint==null){
					//if target is not specified, it moves to unvisited point
					targetPoint = searchNearestUnvisitedPoint(currentEstimatedPoint);
				}
				if(targetPoint!=null){//target is found
					try {
						outputPort.put(targetPoint);
					} catch (PortException e) {
						e.printStackTrace();
					}
					mode = Mode.ACHIEVED;
				}
				else{//move all place in the field but nothing is found
					try {
						outputPort.put(null);
					} catch (PortException e) {
						e.printStackTrace();
					}
					mode = Mode.NOT_ACHIEVED;
				}
				isFinished = true;
			}

			/** Activates the component's service */
			public void activate(){
				super.activate();
			}
			
			/**
			 * Return one of the nearest unvisited points
			 * @param currentP
			 * @return
			 */
			private Point searchNearestUnvisitedPoint(Point currentP){
				Queue<Point> searchQueue = new LinkedList<Point>();
				//ckeckedList: list of points that have already been visited.
				ArrayList<Point> checkedList = new ArrayList<Point>();
				searchQueue.offer(currentP);
				checkedList.add(currentP);
				boolean isFound=false;
				while(searchQueue.size()!=0 && !isFound){
					Point target = searchQueue.poll(); //pull out a point from the queue
//					if(isLocated(target, elemType)){
					if(memorizedMap.get(target)==null){
						System.out.println("\tsize of memorizedMap:\t"+memorizedMap.size());
						isFound = true;
						return(target);
					}
					else{
						//order of inputting into queue is randomly determined
						Point[] tmpArray = new Point[4];
						tmpArray[0] = new Point(target.x, target.y+1);
						tmpArray[1] = new Point(target.x, target.y-1);
						tmpArray[2] = new Point(target.x+1, target.y);
						tmpArray[3] = new Point(target.x-1, target.y);
						
						Point[] pointArray = new Point[4];
						boolean[] isElementSelected = new boolean[4];
						for(int i=0; i<isElementSelected.length;i++){
							isElementSelected[i] = false;
						}
						for(int i=0; i<pointArray.length; i++){
							boolean isAssigned = false;
							while(!isAssigned){
								int randomValue = (int)(Math.random()*4);
								if(!isElementSelected[randomValue]){
									pointArray[i] = tmpArray[randomValue];
									isElementSelected[randomValue] = true;
									isAssigned = true;
								}
							}
						}
						
						//check whether the point is already visited and out of the map
						for(int i=0; i<pointArray.length; i++){
							if(isContainedInMap(pointArray[i])){
								if(!isContaininList(checkedList, pointArray[i])){
									searchQueue.offer(pointArray[i]);
									checkedList.add(pointArray[i]);
								}
							}
						}
					}
				}
				return null;
			}
	}


		/**
		 * Checks whether the specified point is contained in the specified list.
		 * @param l
		 * @param p
		 * @return
		 */
		private boolean isContaininList(List<Point> l, Point p){
			for(Iterator<Point> i = l.iterator();i.hasNext();){
				Point comp_p= (Point)i.next();
				if(comp_p.equals(p)) return true;
			}
			return false;
		}

		/**
		 * Checks whether the specified point is contained in the estimated map.
		 * @return It returns true when the point is within between estimated min and max value of the map.
		 * That is, if the point is actually out of the map but the estimated value is is not collect 
		 * because of the under investigation, it returns true.
		 */
		private boolean isContainedInMap(Point p){
			if(fixed_x_min <= p.x && p.x <= fixed_x_max &&
					fixed_y_min <= p.y && p.y <= fixed_y_max)
			return true;
			else return false;
		}

		/**
		 * Returns the nearest point of the object whose type is specified.
		 * @param cPoint estimated current robot located point
		 * @param oType Type of target object ("Dust" or "ChargeStation")
		 * @return the nearest point of the specified object
		 */
		private Point getNearestPoint(Point cPoint, IdentifiedObject specifiedType){
			Point result = null;
			Set<Point> map_key = memorizedMap.keySet();
			for(Iterator<Point> i= map_key.iterator(); i.hasNext();){
				Point point = (Point)i.next();
				if(memorizedMap.get(point)== specifiedType){
					if(result==null){
						result = point;
					}
					else{
						//choose nearer one
						if((Math.abs(point.x-cPoint.x)+Math.abs(point.y-cPoint.y))<
								(Math.abs(result.x-cPoint.x)+Math.abs(result.y-cPoint.y))){
							result = point;
						}
					}
				}
			}
			return result;
		}



	/**
	 * Move step by step with memorizing the visited point <br/>
	 * @author Hiroyuki NAKAGAWA
	 */
	class MoveWithMemorization extends ComponentBehaviour {
		Point targetPoint; //current target point
		Port<Point> targetPort; //Port
		
		// when this component is activate,  it prints out the component name just once
		boolean forPrint=false;

		public boolean done(){
			return false;
		}

		/**
		 * Constructor
		 */
		public MoveWithMemorization(CleaningRobot agent) {
			super(agent);
			targetPort = new Port<Point>("TargetPoint", PortType.PROVIDED,this);
			this.addPort(targetPort);
			mode = Mode.ACTIVE;
			forPrint = true;
		}

		/**
		 * Execution task
		 */
		public void perform() {
			if(forPrint){
				System.out.println("Component: MoveWithMemorization");
				forPrint = false;
			}
			if(!targetPort.isEmpty()){// target point is changed
				try {
					targetPoint = targetPort.get();
				} catch (PortException e) {
					e.printStackTrace();
				}
//				System.out.println("Move to ..." +"(from"+currentPoint.toString()+" to "+target.toString()+").");
			}
			
			System.out.println("next target point: "+targetPoint);
			int moveResult=1;//1: not move, 0F succeededC-1:failed to move

//			try {
				if(currentEstimatedPoint.x < targetPoint.x && 
						currentEstimatedPoint.x < fixed_x_max){
					moveResult=simulator.moveNext(Direction.EAST);
					if(moveResult==-1){ //fail to move
						fixed_x_max = currentEstimatedPoint.x;
					}
					else currentEstimatedPoint.x += 1;
				}
				else{
					if(targetPoint.x < currentEstimatedPoint.x &&
							fixed_x_min < currentEstimatedPoint.x){
						moveResult=simulator.moveNext(Direction.WEST);
						if(moveResult==-1){ //fail to move
							fixed_x_min = currentEstimatedPoint.x;
						}
						else currentEstimatedPoint.x -= 1;
					}
					else{
						if(currentEstimatedPoint.y<targetPoint.y &&
								currentEstimatedPoint.y < fixed_y_max){
							moveResult=simulator.moveNext(Direction.NORTH);
							if(moveResult==-1){ //fail to move
								fixed_y_max = currentEstimatedPoint.y;
							}
							else currentEstimatedPoint.y += 1;
						}
						else{
							if(targetPoint.y < currentEstimatedPoint.y &&
									fixed_y_min < currentEstimatedPoint.y){
								moveResult=simulator.moveNext(Direction.SOUTH);
								if(moveResult==-1){ //failed to move
									fixed_y_min = currentEstimatedPoint.y;
								}
								else currentEstimatedPoint.y -= 1;
							}
						}
					}
				}
				if(moveResult==0){//move is succeeded
					Point regPoint = new Point(currentEstimatedPoint.x, currentEstimatedPoint.y);
					if(simulator.existsDust()){
						memorizedMap.put(regPoint, IdentifiedObject.DUST);
//						this.mode = Mode.ACHIEVED;
					}
					else{
						if(simulator.isLocatedOnStation()){
							memorizedMap.put(regPoint, IdentifiedObject.STATION);
							//charging on its way
							simulator.chargeBattery();
						}
						else{
							memorizedMap.put(regPoint, IdentifiedObject.NONE);
						}
					}
					if(targetPoint.equals(currentEstimatedPoint)){ 
						//sets the target point again because it arrives at the target point
						System.out.println("Arrived at target point: "+targetPoint);
						targetPoint = null;
						this.mode=Mode.ACHIEVED;
					}
				}
				else{//failed to move
					if(moveResult==-1){
						System.out.println("Move failed to: "+targetPoint);
						targetPoint = null;
						this.mode=Mode.NOT_ACHIEVED;
					}
				}
		}
		
		/** Activates the component's service */
		public void activate(){
			super.activate();
			forPrint = true;
		}
	}
	
}

		


